
#include "dirview.h"

#include <qdir.h>
#include <qmessagebox.h>
#include <qdragobject.h>
#include <qapplication.h>
#include <qheader.h>
#include <qtimer.h>

#include <list>

#define LAYOUT_FILENAME "._"

static const char* folder_closed_xpm[]={
    "16 16 9 1",
    "g c #808080",
    "b c #c0c000",
    "e c #c0c0c0",
    "# c #000000",
    "c c #ffff00",
    ". c None",
    "a c #585858",
    "f c #a0a0a4",
    "d c #ffffff",
    "..###...........",
    ".#abc##.........",
    ".#daabc#####....",
    ".#ddeaabbccc#...",
    ".#dedeeabbbba...",
    ".#edeeeeaaaab#..",
    ".#deeeeeeefe#ba.",
    ".#eeeeeeefef#ba.",
    ".#eeeeeefeff#ba.",
    ".#eeeeefefff#ba.",
    ".##geefeffff#ba.",
    "...##gefffff#ba.",
    ".....##fffff#ba.",
    ".......##fff#b##",
    ".........##f#b##",
    "...........####."};

static const char* folder_open_xpm[]={
    "16 16 11 1",
    "# c #000000",
    "g c #c0c0c0",
    "e c #303030",
    "a c #ffa858",
    "b c #808080",
    "d c #a0a0a4",
    "f c #585858",
    "c c #ffdca8",
    "h c #dcdcdc",
    "i c #ffffff",
    ". c None",
    "....###.........",
    "....#ab##.......",
    "....#acab####...",
    "###.#acccccca#..",
    "#ddefaaaccccca#.",
    "#bdddbaaaacccab#",
    ".eddddbbaaaacab#",
    ".#bddggdbbaaaab#",
    "..edgdggggbbaab#",
    "..#bgggghghdaab#",
    "...ebhggghicfab#",
    "....#edhhiiidab#",
    "......#egiiicfb#",
    "........#egiibb#",
    "..........#egib#",
    "............#ee#"};

static const char * folder_locked[]={
    "16 16 10 1",
    "h c #808080",
    "b c #ffa858",
    "f c #c0c0c0",
    "e c #c05800",
    "# c #000000",
    "c c #ffdca8",
    ". c None",
    "a c #585858",
    "g c #a0a0a4",
    "d c #ffffff",
    "..#a#...........",
    ".#abc####.......",
    ".#daa#eee#......",
    ".#ddf#e##b#.....",
    ".#dfd#e#bcb##...",
    ".#fdccc#daaab#..",
    ".#dfbbbccgfg#ba.",
    ".#ffb#ebbfgg#ba.",
    ".#ffbbe#bggg#ba.",
    ".#fffbbebggg#ba.",
    ".##hf#ebbggg#ba.",
    "...###e#gggg#ba.",
    ".....#e#gggg#ba.",
    "......###ggg#b##",
    ".........##g#b##",
    "...........####."};

static const char * pix_file []={
    "16 16 7 1",
    "# c #000000",
    "b c #ffffff",
    "e c #000000",
    "d c #404000",
    "c c #c0c000",
    "a c #ffffc0",
    ". c None",
    "................",
    ".........#......",
    "......#.#a##....",
    ".....#b#bbba##..",
    "....#b#bbbabbb#.",
    "...#b#bba##bb#..",
    "..#b#abb#bb##...",
    ".#a#aab#bbbab##.",
    "#a#aaa#bcbbbbbb#",
    "#ccdc#bcbbcbbb#.",
    ".##c#bcbbcabb#..",
    "...#acbacbbbe...",
    "..#aaaacaba#....",
    "...##aaaaa#.....",
    ".....##aa#......",
    ".......##......."};

QPixmap *folderLocked = 0;
QPixmap *folderClosed = 0;
QPixmap *folderOpen = 0;
QPixmap *fileNormal = 0;

//****************************************

void Rect2D::bound( const Point2D& p ) {
   if ( empty ) {
      x = p.x;
      y = p.y;
      w = h = 0;
      empty = false;
   }
   else {
      float x2 = x + w;
      float y2 = y + h;
      if ( p.x < x ) x = p.x;
      else if ( x2 < p.x ) x2 = p.x;
      if ( p.y < y ) y = p.y;
      else if ( y2 < p.y ) y2 = p.y;
      w = x2 - x;
      h = y2 - y;
   }
}

void Rect2D::set( const Rect2D& r, float desiredAspectRatio ) {
   (*this) = r;
   if ( !empty ) {
      float a = w/h;
      if ( a > desiredAspectRatio ) {
         // rect is wider than desired
         float newh = h * a / desiredAspectRatio;
         y -= (newh-h)/2;
         h = newh;
      }
      else {
         // rect is taller than desired
         float neww = w * desiredAspectRatio / a;
         x -= (neww-w)/2;
         w = neww;
      }
   }
}

Point2D Rect2D::pixelToPoint( int X, int Y, int W, int H ) const {
   float a = x + w * X/(W-1.0);
   float b = y + h *( 1 - Y/(H-1.0) );
   return Point2D( a, b );
}

/*****************************************************************************
 *
 * Class FileItem
 *
 *****************************************************************************/

void FileItem::setPixmap( const QPixmap *p ) {
   _pixmap = p;
   setup();
   widthChanged( 0 );
   invalidateHeight();
   repaint();
}

const QPixmap *FileItem::pixmap( int i ) const {
   if ( i || !dynamic_cast<DirectoryView*>(listView())->showPixmaps() )
      return 0;
   return _pixmap;
}

/*****************************************************************************
 *
 * Class Directory
 *
 *****************************************************************************/

Directory::Directory( Directory * parent, const QString& filename ) :
   QListViewItem( parent ),
   _file( filename ),
   _showDirsOnly( parent->_showDirsOnly ),
   _pixmap( 0 ),
   _childrenHaveBeenMoved( false )
{
   _parentDir = parent;
   _readable = QDir( fullName() ).isReadable();

   setPixmap( _readable ? folderClosed : folderLocked );
}

Directory::Directory( QListView * parent, const QString& filename ) :
   QListViewItem( parent ),
   _file( filename ),
   _showDirsOnly( ((DirectoryView*)parent)->showDirsOnly() ),
   _pixmap( 0 ),
   _childrenHaveBeenMoved( false )
{
   _parentDir = 0;
   _readable = QDir( fullName() ).isReadable();
}

void Directory::saveState() {

   // tell children to save their states
   QListViewItem * child = firstChild();
   Directory* d;
   for ( ; 0 != child; child = child->nextSibling() )
      if ( 0 != ( d = dynamic_cast< Directory* >( child ) ) )
         d->saveState();

   if ( _childrenHaveBeenMoved ) {
      QString buff = fullName()+LAYOUT_FILENAME;
      QFile file( buff );
      file.open( IO_WriteOnly );
      for ( child = firstChild(); 0 != child; child = child->nextSibling() ) {
         buff.truncate( 0 );
         if ( 0 != ( d = dynamic_cast< Directory* >( child ) ) ) {
            buff.sprintf( "%s %f %f\n", d->text(0).latin1(),
               d->getPoint().x, d->getPoint().y
            );
         }
         else if ( FileItem* f = dynamic_cast< FileItem* >( child ) ) {
            buff.sprintf( "%s %f %f\n", f->text(0).latin1(), 
               f->getPoint().x, f->getPoint().y 
            );
         }
         file.writeBlock( buff.latin1(), buff.length() );
      }
      file.close();
   }
}

void Directory::setPixmap( const QPixmap *px ) {
   _pixmap = px;
   setup();
   widthChanged( 0 );
   invalidateHeight();
   repaint();
}

const QPixmap *Directory::pixmap( int i ) const {
   if ( i || !dynamic_cast<DirectoryView*>(listView())->showPixmaps() )
      return 0;
   return _pixmap;
}

void Directory::readContents() {
   QString s( fullName() );
   QDir thisDir( s );
   if ( !thisDir.isReadable() ) {
      _readable = FALSE;
      setExpandable( false );
      return;
   }

   listView()->setUpdatesEnabled( false );
   const QFileInfoList * files = thisDir.entryInfoList();
   if ( files ) {
      QFileInfoListIterator it( *files );
      QFileInfo * fi;
      while( (fi=it.current()) != 0 ) {
         ++it;
         if ( fi->fileName() == "." || fi->fileName() == ".." )
            ; // nothing
         else if ( fi->isSymLink() && !_showDirsOnly ) {
            FileItem *item = new FileItem( this, fi->fileName(),
               "Symbolic Link", "" );
            item->setPixmap( fileNormal );
         }
         else if ( fi->isDir() )
            (void)new Directory( this, fi->fileName() );
         else if ( !_showDirsOnly ) {
            if ( fi->fileName() != LAYOUT_FILENAME ) {
               QString sizeString;
               // Padding with spaces is necessary so that, if the user
               // requests that the files be sorted by "size", the ordering
               // of strings will be correct.
               sizeString.sprintf( "%9u", fi->size() );
               FileItem *item = new FileItem(
                  this, fi->fileName(), fi->isFile() ? "File" : "Special",
                  sizeString
               );
               item->setPixmap( fileNormal );
            }
         }
      }

      QFileInfo layoutFileInfo( s + LAYOUT_FILENAME );

      if ( layoutFileInfo.exists() ) {
         // build a list of children who haven't been arranged
         list< QListViewItem * > unarrangedChildren;
         QListViewItem * child = firstChild();
         for ( ; 0 != child; child = child->nextSibling() ) {
            unarrangedChildren.push_back( child );
         }

         // open the layout file for reading
         QFile layoutFile( s + LAYOUT_FILENAME );
         layoutFile.open( IO_ReadOnly );
         QString line;
         while ( layoutFile.readLine( line, 500 ) != -1 ) {
            QStringList lst( QStringList::split( " ", line ) );
            if ( lst.count() == 3 ) {
               list< QListViewItem * >::iterator it
                  = unarrangedChildren.begin();
               for ( ; it != unarrangedChildren.end(); ++it ) {
                  if ( (*it)->text(0) == lst[0] ) {
                     if ( Directory* d = dynamic_cast<Directory*>(*it) ) {
                        d->_p = Point2D( lst[1].toFloat(), lst[2].toFloat() );
                        _r.bound( d->_p );
                     }
                     else if ( FileItem* f = dynamic_cast<FileItem*>(*it) ) {
                        f->_p = Point2D( lst[1].toFloat(), lst[2].toFloat() );
                        _r.bound( f->_p );
                     }
                     unarrangedChildren.erase( it );
                     break;
                  }
               }
            }
         }
         layoutFile.close();

         // auto-arrange all remaining children
         float x = _r.x + _r.w/2;
         float y = _r.y - 6;
         int dirCount = 0, fileCount = 0;
         list< QListViewItem * >::iterator it = unarrangedChildren.begin();
         for ( ; it != unarrangedChildren.end(); ++it ) {
            if ( Directory* d = dynamic_cast<Directory*>(*it) ) {
               d->_p = Point2D( x-5, y - dirCount * 4 );
               _r.bound( d->_p );
               ++dirCount;
            }
            else if ( FileItem* f = dynamic_cast<FileItem*>(*it) ) {
               f->_p = Point2D( x+5, y - fileCount * 4 );
               _r.bound( f->_p );
               ++fileCount;
            }
         }
      }
      else {
         // generate a default layout
         int dirCount = 0, fileCount = 0;
         QListViewItem * child = firstChild();
         for ( ; 0 != child; child = child->nextSibling() ) {
            if ( Directory* d = dynamic_cast< Directory* >( child ) ) {
               d->_p = Point2D( 0, dirCount * 4 );
               _r.bound( d->_p );
               ++dirCount;
            }
            else if ( FileItem* f = dynamic_cast< FileItem* >( child ) ) {
               f->_p = Point2D( 10, fileCount * 4 );
               _r.bound( f->_p );
               ++fileCount;
            }
         }
      }

      // expand the rectangle
      _r.x -= 3;
      _r.y -= 3;
      _r.w += 6;
      _r.h += 6;
   }
   listView()->setUpdatesEnabled( true );
}

void Directory::setOpen( bool flag ) {
   setPixmap( flag ? folderOpen : folderClosed );

   if ( flag && 0 == childCount() ) {
      readContents();
   }
   QListViewItem::setOpen( flag );
}

void Directory::setup() {
   setExpandable( true );
   QListViewItem::setup();
}

QString Directory::fullName() {
   QString s;
   if ( _parentDir ) {
      s = _parentDir->fullName();
      s.append( _file.name() );
      s.append( "/" );
   } else {
      s = _file.name();
   }
   return s;
}

QString Directory::text( int column ) const {
   switch ( column ) {
   case 0 : return _file.name();
      break;
   case 1 : return _readable ? "Directory" : "Unreadable Directory";
      break;
   default :
      return QListViewItem::text( column );
   }
}

/*****************************************************************************
 *
 * Class DirectoryView
 *
 *****************************************************************************/

DirectoryView::DirectoryView( QWidget *parent, const char *name, bool sdo ) :
   QListView( parent, name ), _app( 0 ), dirsOnly( sdo ), oldCurrent( 0 ),
   dropItem( 0 ), mousePressed( FALSE ), _showPixmaps( true )
{
    autoopen_timer = new QTimer( this );
    if ( !folderLocked ) {
	folderLocked = new QPixmap( folder_locked );
	folderClosed = new QPixmap( folder_closed_xpm );
	folderOpen = new QPixmap( folder_open_xpm );
	fileNormal = new QPixmap( pix_file );
    }

    connect( this, SIGNAL( doubleClicked( QListViewItem * ) ),
	     this, SLOT( slotFolderSelected( QListViewItem * ) ) );
    connect( this, SIGNAL( returnPressed( QListViewItem * ) ),
	     this, SLOT( slotFolderSelected( QListViewItem * ) ) );

    setAcceptDrops( TRUE );
    viewport()->setAcceptDrops( TRUE );

    connect( autoopen_timer, SIGNAL( timeout() ),
	     this, SLOT( openFolder() ) );
}

void DirectoryView::loadSubtree( const QString & s ) {
   saveState();
   emit contentsCleared();
   clear();
   QDir dir( s );
   dir.convertToAbs();
   QString s2;
   if ( s == "/" ) s2 = s;
   else s2 = dir.path() + "/";
   Directory * root = new Directory( this, s2 );
   root->setOpen( TRUE );
}

void DirectoryView::saveState() const {
   QListViewItem * foo = firstChild();
   if ( 0 != foo ) {
      Directory * dir = dynamic_cast< Directory* >( foo );
      dir->saveState();
   }
}

void DirectoryView::slotFolderSelected( QListViewItem *i )
{
    if ( !i || !showDirsOnly() )
	return;
    Directory *dir = (Directory*)i;
    emit folderSelected( dir->fullName() );
}

void DirectoryView::openFolder()
{
    autoopen_timer->stop();
    if ( dropItem && !dropItem->isOpen() ) {
	dropItem->setOpen( TRUE );
	dropItem->repaint();
    }
}

static const int autoopenTime = 750;


void DirectoryView::contentsDragEnterEvent( QDragEnterEvent *e )
{
    if ( !QUriDrag::canDecode(e) ) {
	e->ignore();
	return;
    }

    oldCurrent = currentItem();

    QListViewItem *i = itemAt( contentsToViewport(e->pos()) );
    if ( i ) {
	dropItem = i;
	autoopen_timer->start( autoopenTime );
    }
}


void DirectoryView::contentsDragMoveEvent( QDragMoveEvent *e )
{
    if ( !QUriDrag::canDecode(e) ) {
	e->ignore();
	return;
    }

    QPoint vp = contentsToViewport( ( (QDragMoveEvent*)e )->pos() );
    QListViewItem *i = itemAt( vp );
    if ( i ) {
	setSelected( i, TRUE );
	e->accept();
	if ( i != dropItem ) {
	    autoopen_timer->stop();
	    dropItem = i;
	    autoopen_timer->start( autoopenTime );
	}
	switch ( e->action() ) {
	case QDropEvent::Copy:
	    break;
	case QDropEvent::Move:
	    e->acceptAction();
	    break;
	case QDropEvent::Link:
	    e->acceptAction();
	    break;
	default:
	    ;
	}
    } else {
	e->ignore();
	autoopen_timer->stop();
	dropItem = 0;
    }
}

void DirectoryView::contentsDragLeaveEvent( QDragLeaveEvent * )
{
    autoopen_timer->stop();
    dropItem = 0;

    setCurrentItem( oldCurrent );
    setSelected( oldCurrent, TRUE );
}

void DirectoryView::contentsDropEvent( QDropEvent *e )
{
    autoopen_timer->stop();

    if ( !QUriDrag::canDecode(e) ) {
	e->ignore();
	return;
    }

    QListViewItem *item = itemAt( contentsToViewport(e->pos()) );
    if ( item ) {

	QStrList lst;

	QUriDrag::decode( e, lst );

	QString str;

	switch ( e->action() ) {
	    case QDropEvent::Copy:
	    str = "Copy";
	    break;
	    case QDropEvent::Move:
	    str = "Move";
	    e->acceptAction();
	    break;
	    case QDropEvent::Link:
	    str = "Link";
	    e->acceptAction();
	    break;
	    default:
	    str = "Unknown";
	}

	str += "\n\n";

	e->accept();

	for ( uint i = 0; i < lst.count(); ++i ) {
	    QString filename = lst.at( i );
	    str += filename + "\n";
	}
	str += QString( "\nTo\n\n   %1" )
	       .arg( fullPath(item) );

	QMessageBox::information( this, "Drop target", str, "Not implemented" );
    }
    else e->ignore();
}


QString DirectoryView::fullPath(QListViewItem* item)
{
    QString fullpath = item->text(0);
    while ( (item=item->parent()) ) {
	if ( item->parent() )
	    fullpath = item->text(0) + "/" + fullpath;
	else
	    fullpath = item->text(0) + fullpath;
    }
    return fullpath;
}

void DirectoryView::contentsMousePressEvent( QMouseEvent* e )
{
    QListView::contentsMousePressEvent(e);
    QPoint p( contentsToViewport( e->pos() ) );
    QListViewItem *i = itemAt( p );
    if ( i ) {
	// if the user clicked into the root decoration of the item, don't try to start a drag!
	if ( p.x() > header()->cellPos( header()->mapToActual( 0 ) ) +
	     treeStepSize() * ( i->depth() + ( rootIsDecorated() ? 1 : 0) ) + itemMargin() ||
	     p.x() < header()->cellPos( header()->mapToActual( 0 ) ) ) {
	    presspos = e->pos();
	    mousePressed = TRUE;
	}
    }
}

void DirectoryView::contentsMouseMoveEvent( QMouseEvent* e )
{
    if ( mousePressed && ( presspos - e->pos() ).manhattanLength() > QApplication::startDragDistance() ) {
	mousePressed = FALSE;
	QListViewItem *item = itemAt( contentsToViewport(presspos) );
	if ( item ) {
	    QString source = fullPath(item);
	    if ( QFile::exists(source) ) {
		QUriDrag* ud = new QUriDrag(viewport());
		ud->setUnicodeUris( source );
		if ( ud->drag() )
		    QMessageBox::information( this, "Drag source",
					      QString("Delete ")+source, "Not implemented" );
	    }
	}
    }
}

void DirectoryView::contentsMouseReleaseEvent( QMouseEvent * ) {
   mousePressed = FALSE;
}

//void DirectoryView::setDir( const QString &s )
//{
//    QListViewItemIterator it( this );
//    ++it;
//    for ( ; it.current(); ++it ) {
//        it.current()->setOpen( FALSE );
//    }
//
//    QStringList lst( QStringList::split( "/", s ) );
//    QListViewItem *item = firstChild();
//    QStringList::Iterator it2 = lst.begin();
//    for ( ; it2 != lst.end(); ++it2 ) {
//        while ( item ) {
//            if ( item->text( 0 ) == *it2 ) {
//                item->setOpen( TRUE );
//                break;
//            }
//            item = item->itemBelow();
//        }
//    }
//
//    if ( item )
//        setCurrentItem( item );
//}

void DirectoryView::toggleIcons() {  // FIXME: doesn't cause proper update
   _showPixmaps = ! _showPixmaps;
   // triggerUpdate(); // FIXME: should we try calling this ?
   repaint();
}

void DirectoryView::loadParentOfCurrentSubtree() {
   QListViewItem * foo = firstChild();
   if ( 0 != foo ) {
      Directory * dir = dynamic_cast< Directory* >( foo );
      QDir d( dir->fullName() );
      d.convertToAbs();
      d.cdUp();
      loadSubtree( d.path() );
   }
}

void DirectoryView::loadSubtreeUnderCurrentItem() {
   QListViewItem * foo = currentItem();
   if ( 0 != foo ) {
      Directory * dir = dynamic_cast< Directory* >( foo );
      if ( 0 == dir )
         dir = dynamic_cast< Directory* >( foo->parent() );
      QString s = dir->fullName();
      loadSubtree( s );
   }
}

void DirectoryView::computeSizeOfCurrentDirectory() {
   QListViewItem * foo = currentItem();
   if ( 0 != foo ) {
      Directory * dir = dynamic_cast< Directory* >( foo );
      if ( 0 == dir )
         dir = dynamic_cast< Directory* >( foo->parent() );
      QString s = dir->fullName();

      // At this point we run "du" to find out how much disk space
      // is used by the directory.  We use system() to run it.
      // The only information that system() returns is an exit code;
      // hence we choose the sh command so that its exit code will
      // correspond to the value computed by "du" :-)
      //
      QString str;
      // (Note that we don't detect/handle errors well yet, so the
      // redirection of stderr to null is probably not a good idea.
      //str.sprintf( "exit `du -bs %s | cut -f 1` &> /dev/null", s.latin1() );
      str.sprintf( "exit `du -bs %s | cut -f 1`", s.latin1() );

      int exitCode = system( str.latin1() );

      // FIXME: this doesn't work !
      // Let foo be some file.  Then, experimentation shows that
      //    system( "echo `/usr/bin/du -bs ./foo | /usr/bin/cut -f 1`" );
      // echos the correct value (but of course has a useless exit value),
      // and
      //    system( "exit `/bin/echo 20 30 | /usr/bin/cut -f 1 -d ' '`" );
      // returns the expected value (i.e. 20 multiplied by 256),
      // *but*
      //    system( "exit `/usr/bin/du -bs ./foo | /usr/bin/cut -f 1`" );
      // returns 0 for some mysterious reason.

      // The docs say that system should return -1 if there's an error,
      // or the exit code of the sh command otherwise.
      // Unfortunately, experimentation with Redhat 7 shows that 256
      // is returned if there's an error, otherwise the sh command's
      // error code is returned multiplied by 256.

      if ( exitCode == 256 || exitCode == -1 || (exitCode & 0xff) != 0 ) {
         dir->setText( 2, "?" );
      }
      else {
         s.setNum( exitCode/256 );
         dir->setText( 2, s );
      }
   }
}

void DirectoryView::causeQuit() {
   saveState();
   _app->quit();
}

